/*
 * GStreamer
 * Copyright 2007 Ole Andr Vadla Ravns <ole.andre.ravnas@tandberg.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-osxvideosrc
 *
 * <refsect2>
 * osxvideosrc can be used to capture video from capture devices on OS X.
 * <title>Example launch line</title>
 * <para>
 * <programlisting>
 * gst-launch osxvideosrc ! osxvideosink
 * </programlisting>
 * This pipeline shows the video captured from the default capture device.
 * </para>
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include "gstosxvideosrc.h"

#define WIDTH 640
#define HEIGHT 480
#define FRAMERATE 30

GST_DEBUG_CATEGORY_STATIC (gst_osxvideosrc_debug);
#define GST_CAT_DEFAULT gst_osxvideosrc_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
};

static GstStaticPadTemplate src_template =
  GST_STATIC_PAD_TEMPLATE (
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("video/x-raw-yuv, "
        "format = (fourcc) YUY2, "
        "width = (int) " G_STRINGIFY (WIDTH)  ", "
        "height = (int) " G_STRINGIFY (HEIGHT) ", "
        "framerate = (fraction) " G_STRINGIFY (FRAMERATE) "/1")
  );

GST_BOILERPLATE (GstOSXVideoSrc, gst_osxvideosrc, GstPushSrc,
    GST_TYPE_PUSH_SRC);

static void gst_osxvideosrc_dispose (GObject * object);
static void gst_osxvideosrc_finalize (GstOSXVideoSrc * osxvideosrc);

static gboolean gst_osxvideosrc_start (GstBaseSrc * src);
static gboolean gst_osxvideosrc_stop (GstBaseSrc * src);
static GstFlowReturn gst_osxvideosrc_create (GstPushSrc * src, GstBuffer ** buf);

static GstStateChangeReturn gst_osxvideosrc_change_state (GstElement * element, GstStateChange transition);

static void
gst_osxvideosrc_base_init (gpointer gclass)
{
  static GstElementDetails element_details = {
    "Video Source (OSX)",
    "Source/Video",
    "Reads raw frames from a capture device on OS X",
    "Ole Andre Vadla Ravnaas <ole.andre.ravnas@tandberg.com>"
  };
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  GST_DEBUG (G_STRFUNC);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_template));
  gst_element_class_set_details (element_class, &element_details);
}

/* initialize the plugin's class */
static void
gst_osxvideosrc_class_init (GstOSXVideoSrcClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *element_class;
  GstBaseSrcClass *basesrc_class;
  GstPushSrcClass *pushsrc_class;

  GST_DEBUG (G_STRFUNC);

  gobject_class = G_OBJECT_CLASS (klass);
  element_class = GST_ELEMENT_CLASS (klass);
  basesrc_class = GST_BASE_SRC_CLASS (klass);
  pushsrc_class = GST_PUSH_SRC_CLASS (klass);

  gobject_class->dispose = gst_osxvideosrc_dispose;
  gobject_class->finalize = (GObjectFinalizeFunc) gst_osxvideosrc_finalize;

  element_class->change_state = gst_osxvideosrc_change_state;

  basesrc_class->start = gst_osxvideosrc_start;
  basesrc_class->stop = gst_osxvideosrc_stop;

  pushsrc_class->create = gst_osxvideosrc_create;
}

static gboolean init_basics (GstOSXVideoSrc * self);

static void
gst_osxvideosrc_init (GstOSXVideoSrc * self, GstOSXVideoSrcClass * klass)
{
  GstPad * srcpad = GST_BASE_SRC_PAD (self);

  GST_DEBUG_OBJECT (self, G_STRFUNC);

  gst_base_src_set_format (GST_BASE_SRC (self), GST_FORMAT_TIME);
  gst_base_src_set_live (GST_BASE_SRC (self), TRUE);

  gst_pad_use_fixed_caps (srcpad);
  gst_pad_set_caps (srcpad, gst_caps_copy (
        gst_pad_get_pad_template_caps (srcpad)));

  self->seq = 0;

  if (!init_basics (self))
    return;
}

static void
gst_osxvideosrc_dispose (GObject * object)
{
  GST_DEBUG_OBJECT (object, G_STRFUNC);

  G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
gst_osxvideosrc_finalize (GstOSXVideoSrc * self)
{
  GST_DEBUG_OBJECT (self, G_STRFUNC);

  G_OBJECT_CLASS (parent_class)->finalize (G_OBJECT (self));
}

static gboolean
init_basics (GstOSXVideoSrc * self)
{
  ComponentResult err = noErr;
  DigitizerInfo vd_info;
  long msPerFrame;
  Fixed framesPerSecond;
  long bytesPerSecond;
  Rect rect;
  /*ImageDescriptionHandle desc;*/

  self->seq_grab = OpenDefaultComponent (SeqGrabComponentType, 0);
  if (self->seq_grab == NULL) {
    GST_ERROR ("OpenDefaultComponent failed");
    err = paramErr;
    goto OUT;
  }

  err = SGInitialize (self->seq_grab);
  if (err != noErr) {
    GST_ERROR ("SGInitialize failed");
    goto OUT;
  }

  err = SGSetDataRef (self->seq_grab, 0, 0, seqGrabDontMakeMovie);
  if (err != noErr) {
    GST_ERROR ("SGSetDataRef failed");
    goto OUT;
  }

  err = SGNewChannel (self->seq_grab, VideoMediaType, &self->video_chan);
  if (err != noErr) {
    GST_ERROR ("SGNewChannel failed");
    goto OUT;
  }

  self->vd = SGGetVideoDigitizerComponent (self->video_chan);

  err = VDGetDigitizerInfo (self->vd, &vd_info);
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDGetDigitizerInfo failed");
    goto OUT;
  }

  GST_DEBUG_OBJECT (self, "min=(%d, %d), max=(%d, %d)",
      vd_info.minDestWidth, vd_info.minDestHeight,
      vd_info.maxDestWidth, vd_info.maxDestHeight);

  if ((vd_info.outputCurrentFlags & digiOutDoesCompress) == 0)
  {
    GST_ERROR_OBJECT (self, "Device doesn't support compressed output");
    err = paramErr;
    goto OUT;
  }

  /*
  if ((vd_info.outputCurrentFlags & digiOutDoesAsyncGrabs) == 0)
  {
    GST_ERROR_OBJECT (self, "Device doesn't support async grabbing");
    err = paramErr;
    goto OUT;
  }
  */

  err = VDGetDataRate (self->vd, &msPerFrame, &framesPerSecond,
      &bytesPerSecond);
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDGetDataRate failed");
    goto OUT;
  }

  GST_DEBUG_OBJECT (self, "msPerFrame=%ld, framesPerSecond=%f, bytesPerSecond=%ld",
      msPerFrame, FixedToFloat (framesPerSecond), bytesPerSecond);

  /* FIXME: don't hardcode the caps */
  /*g_assert (FixedToInt (framesPerSecond) == FRAMERATE);*/

  err = VDSetCompressionOnOff (self->vd, TRUE);
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDSetCompressionOnOff(TRUE) failed");
    goto OUT;
  }

  /*
  err = VDSetTimeBase (self->vd, NewTimeBase ());
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDSetTimeBase failed");
    goto OUT;
  }*/

  err = VDSetFrameRate (self->vd, IntToFixed (FRAMERATE));
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDSetFrameRate failed");
    goto OUT;
  }

  /*
  err = VDSetDataRate (self->vd, 614400 * FRAMERATE);
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDSetDataRate failed");
    goto OUT;
  }
  */

  rect.top = 0;
  rect.left = 0;
  rect.right = 640;
  rect.bottom = 480;

  err = VDSetCompression (self->vd, 0, 0, &rect, 0, 0, 0);
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDSetCompression failed");
    goto OUT;
  }

  /*
  desc = (ImageDescriptionHandle) NewHandle (4);
  err = VDGetImageDescription (self->vd, desc);
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDGetImageDescription failed");
    goto OUT;
  }

  GST_DEBUG_OBJECT (self, "width=%d, height=%d", (*desc)->width, (*desc)->height);
  */

OUT:
  return (err == noErr);
}

static gboolean
gst_osxvideosrc_start (GstBaseSrc * src)
{
  GstOSXVideoSrc * self = GST_OSXVIDEOSRC (src);
  ComponentResult err = noErr;

  GST_DEBUG_OBJECT (src, "entering");

  err = VDCompressOneFrameAsync (self->vd);
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDCompressOneFrameAsync failed with error %ld",
        err);
    goto OUT;
  }

  GST_DEBUG_OBJECT (self, "wee"); 

OUT:
  return (err == noErr);
}

static gboolean
gst_osxvideosrc_stop (GstBaseSrc * src)
{
  GST_DEBUG_OBJECT (src, "FIXME");

  return TRUE;
}

static GstFlowReturn
gst_osxvideosrc_create (GstPushSrc * src, GstBuffer ** buf)
{
  GstFlowReturn ret = GST_FLOW_UNEXPECTED;
  GstOSXVideoSrc * self = GST_OSXVIDEOSRC (src);
  ComponentResult err;
  UInt8 count = 0;
  Ptr data = NULL;
  long data_size;
  UInt8 similarity;
  TimeRecord t;
  GstCaps * caps;

  /*GST_DEBUG_OBJECT (self, G_STRFUNC);*/

  do
  {
    err = VDCompressDone (self->vd, &count, &data, &data_size, &similarity, &t);
    if (err != noErr)
    {
      GST_ERROR_OBJECT (self, "VDCompressDone failed with error %ld", err);
      goto OUT;
    }

    if (count == 0)
      sleep (0);
  }
  while (count == 0);

  /* HACK, sort of: discard old frames */
  while (count > 1)
  {
    err = VDReleaseCompressBuffer (self->vd, data);
    if (err != noErr)
    {
      GST_ERROR_OBJECT (self, "VDReleaseCompressBuffer failed with error %ld", err);
      ret = GST_FLOW_UNEXPECTED;
    }

    err = VDCompressDone (self->vd, &count, &data, &data_size, &similarity, &t);
    if (err != noErr)
    {
      GST_ERROR_OBJECT (self, "VDCompressDone failed with error %ld", err);
      goto OUT;
    }
  }

  /*
  err = VDCompressOneFrameAsync (self->vd);
  if (err != noErr)
  {
    GST_ERROR_OBJECT (self, "VDCompressOneFrameAsync failed with error %ld",
        err);
    goto OUT;
  }*/

  /*GST_DEBUG_OBJECT (self, "count=%d, data_size=%ld", count, data_size);*/

  *buf = gst_buffer_new_and_alloc (data_size);
  GST_BUFFER_OFFSET (*buf) = self->seq;
  GST_BUFFER_TIMESTAMP (*buf) = gst_util_uint64_scale_int (
      self->seq, GST_SECOND, FRAMERATE);
  GST_BUFFER_DURATION (*buf) = gst_util_uint64_scale_int (
      1, GST_SECOND, FRAMERATE);
  memcpy (GST_BUFFER_DATA (*buf), data, data_size);

  caps = gst_pad_get_caps (GST_BASE_SRC_PAD (src));
  gst_buffer_set_caps (*buf, caps);
  gst_caps_unref (caps);

  self->seq++;

  ret = GST_FLOW_OK;

OUT:
  if (data != NULL)
  {
    err = VDReleaseCompressBuffer (self->vd, data);
    if (err != noErr)
    {
      GST_ERROR_OBJECT (self, "VDReleaseCompressBuffer failed with error %ld", err);
      ret = GST_FLOW_UNEXPECTED;
    }
  }

  return GST_FLOW_OK;
}

static GstStateChangeReturn
gst_osxvideosrc_change_state (GstElement * element, GstStateChange transition)
{
  GstOSXVideoSrc * self = GST_OSXVIDEOSRC (element);
  GstStateChangeReturn ret = GST_STATE_CHANGE_FAILURE;
  ComponentResult err;

  switch (transition)
  {
    case GST_STATE_CHANGE_NULL_TO_READY:
      GST_DEBUG_OBJECT (element, "GST_STATE_CHANGE_NULL_TO_READY");
      break;
    case GST_STATE_CHANGE_READY_TO_PAUSED:
      GST_DEBUG_OBJECT (element, "GST_STATE_CHANGE_READY_TO_PAUSED");
      break;
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
      GST_DEBUG_OBJECT (element, "GST_STATE_CHANGE_PAUSED_TO_PLAYING");

      err = VDSetCompressionOnOff (self->vd, TRUE);
      if (err != noErr)
      {
        GST_ERROR_OBJECT (self, "VDSetCompressionOnOff(TRUE) failed");
        ret = GST_STATE_CHANGE_FAILURE;
      }

      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      GST_DEBUG_OBJECT (element, "GST_STATE_CHANGE_READY_TO_NULL");
      break;
    case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
      GST_DEBUG_OBJECT (element, "GST_STATE_CHANGE_PLAYING_TO_PAUSED");

      /* TODO: is this enough? */
      err = VDSetCompressionOnOff (self->vd, FALSE);
      if (err != noErr)
      {
        GST_ERROR_OBJECT (self, "VDSetCompressionOnOff(FALSE) failed");
        ret = GST_STATE_CHANGE_FAILURE;
      }

      break;
    case GST_STATE_CHANGE_PAUSED_TO_READY:
      GST_DEBUG_OBJECT (element, "GST_STATE_CHANGE_PAUSED_TO_READY");
      break;
    default:
      break;
  }

  return ret;
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  GST_DEBUG (G_STRFUNC);

  GST_DEBUG_CATEGORY_INIT (gst_osxvideosrc_debug, "osxvideosrc",
      0, "Reads raw frames from a capture device on OS X");

  return gst_element_register (plugin, "osxvideosrc",
      GST_RANK_NONE, GST_TYPE_OSXVIDEOSRC);
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "osxvideosrc",
    "Reads raw frames from a capture device on OS X",
    plugin_init, VERSION, "LGPL", "GStreamer", "http://gstreamer.net/")
